@@ -17,7 +17,6 @@ import (
 	"io"
 	"math/big"
 
-	"boringssl.googlesource.com/boringssl/ssl/test/runner/hrss"
 	"golang.org/x/crypto/curve25519"
 )
 
@@ -231,28 +230,29 @@ func (ka *rsaKeyAgreement) peerSignatureAlgorithm() signatureAlgorithm {
 	return 0
 }
 
-// A ecdhCurve is an instance of ECDH-style key agreement for TLS.
-type ecdhCurve interface {
-	// offer generates a keypair using rand. It returns the encoded |publicKey|.
-	offer(rand io.Reader) (publicKey []byte, err error)
+// A kemImplementation is an instance of KEM-style construction for TLS.
+type kemImplementation interface {
+	// generate generates a keypair using rand. It returns the encoded public key.
+	generate(rand io.Reader) (publicKey []byte, err error)
 
-	// accept responds to the |peerKey| generated by |offer| with the acceptor's
-	// |publicKey|, and returns agreed-upon |preMasterSecret| to the acceptor.
-	accept(rand io.Reader, peerKey []byte) (publicKey []byte, preMasterSecret []byte, err error)
+	// encap generates a symmetric, shared secret, encapsulates it with |peerKey|.
+	// It returns the encapsulated shared secret and the secret itself.
+	encap(rand io.Reader, peerKey []byte) (ciphertext []byte, secret []byte, err error)
 
-	// finish returns the computed |preMasterSecret|, given the |peerKey|
-	// generated by |accept|.
-	finish(peerKey []byte) (preMasterSecret []byte, err error)
+	// decap decapsulates |ciphertext| and returns the resulting shared secret.
+	decap(ciphertext []byte) (secret []byte, err error)
 }
 
-// ellipticECDHCurve implements ecdhCurve with an elliptic.Curve.
-type ellipticECDHCurve struct {
+// ecdhKEM implements kemImplementation with an elliptic.Curve.
+//
+// TODO(davidben): Move this to Go's crypto/ecdh.
+type ecdhKEM struct {
 	curve          elliptic.Curve
 	privateKey     []byte
 	sendCompressed bool
 }
 
-func (e *ellipticECDHCurve) offer(rand io.Reader) (publicKey []byte, err error) {
+func (e *ecdhKEM) generate(rand io.Reader) (publicKey []byte, err error) {
 	var x, y *big.Int
 	e.privateKey, x, y, err = elliptic.GenerateKey(e.curve, rand)
 	if err != nil {
@@ -269,38 +269,37 @@ func (e *ellipticECDHCurve) offer(rand io.Reader) (publicKey []byte, err error)
 	return ret, nil
 }
 
-func (e *ellipticECDHCurve) accept(rand io.Reader, peerKey []byte) (publicKey []byte, preMasterSecret []byte, err error) {
-	publicKey, err = e.offer(rand)
+func (e *ecdhKEM) encap(rand io.Reader, peerKey []byte) (ciphertext []byte, secret []byte, err error) {
+	ciphertext, err = e.generate(rand)
 	if err != nil {
 		return nil, nil, err
 	}
-	preMasterSecret, err = e.finish(peerKey)
+	secret, err = e.decap(peerKey)
 	if err != nil {
 		return nil, nil, err
 	}
 	return
 }
 
-func (e *ellipticECDHCurve) finish(peerKey []byte) (preMasterSecret []byte, err error) {
-	x, y := elliptic.Unmarshal(e.curve, peerKey)
+func (e *ecdhKEM) decap(ciphertext []byte) (secret []byte, err error) {
+	x, y := elliptic.Unmarshal(e.curve, ciphertext)
 	if x == nil {
 		return nil, errors.New("tls: invalid peer key")
 	}
 	x, _ = e.curve.ScalarMult(x, y, e.privateKey)
-	preMasterSecret = make([]byte, (e.curve.Params().BitSize+7)>>3)
+	secret = make([]byte, (e.curve.Params().BitSize+7)>>3)
 	xBytes := x.Bytes()
-	copy(preMasterSecret[len(preMasterSecret)-len(xBytes):], xBytes)
-
-	return preMasterSecret, nil
+	copy(secret[len(secret)-len(xBytes):], xBytes)
+	return secret, nil
 }
 
-// x25519ECDHCurve implements ecdhCurve with X25519.
-type x25519ECDHCurve struct {
+// x25519KEM implements kemImplementation with X25519.
+type x25519KEM struct {
 	privateKey [32]byte
 	setHighBit bool
 }
 
-func (e *x25519ECDHCurve) offer(rand io.Reader) (publicKey []byte, err error) {
+func (e *x25519KEM) generate(rand io.Reader) (publicKey []byte, err error) {
 	_, err = io.ReadFull(rand, e.privateKey[:])
 	if err != nil {
 		return
@@ -313,25 +312,24 @@ func (e *x25519ECDHCurve) offer(rand io.Reader) (publicKey []byte, err error) {
 	return out[:], nil
 }
 
-func (e *x25519ECDHCurve) accept(rand io.Reader, peerKey []byte) (publicKey []byte, preMasterSecret []byte, err error) {
-	publicKey, err = e.offer(rand)
+func (e *x25519KEM) encap(rand io.Reader, peerKey []byte) (ciphertext []byte, secret []byte, err error) {
+	ciphertext, err = e.generate(rand)
 	if err != nil {
 		return nil, nil, err
 	}
-	preMasterSecret, err = e.finish(peerKey)
+	secret, err = e.decap(peerKey)
 	if err != nil {
 		return nil, nil, err
 	}
 	return
 }
 
-func (e *x25519ECDHCurve) finish(peerKey []byte) (preMasterSecret []byte, err error) {
-	if len(peerKey) != 32 {
+func (e *x25519KEM) decap(ciphertext []byte) (secret []byte, err error) {
+	if len(ciphertext) != 32 {
 		return nil, errors.New("tls: invalid peer key")
 	}
-	var out, peerKeyCopy [32]byte
-	copy(peerKeyCopy[:], peerKey)
-	curve25519.ScalarMult(&out, &e.privateKey, &peerKeyCopy)
+	var out [32]byte
+	curve25519.ScalarMult(&out, &e.privateKey, (*[32]byte)(ciphertext))
 
 	// Per RFC 7748, reject the all-zero value in constant time.
 	var zeros [32]byte
@@ -342,104 +340,18 @@ func (e *x25519ECDHCurve) finish(peerKey []byte) (preMasterSecret []byte, err er
 	return out[:], nil
 }
 
-// cecpq2Curve implements CECPQ2, which is HRSS+SXY combined with X25519.
-type cecpq2Curve struct {
-	x25519PrivateKey [32]byte
-	hrssPrivateKey   hrss.PrivateKey
-}
-
-func (e *cecpq2Curve) offer(rand io.Reader) (publicKey []byte, err error) {
-	if _, err := io.ReadFull(rand, e.x25519PrivateKey[:]); err != nil {
-		return nil, err
-	}
-
-	var x25519Public [32]byte
-	curve25519.ScalarBaseMult(&x25519Public, &e.x25519PrivateKey)
-
-	e.hrssPrivateKey = hrss.GenerateKey(rand)
-	hrssPublic := e.hrssPrivateKey.PublicKey.Marshal()
-
-	var ret []byte
-	ret = append(ret, x25519Public[:]...)
-	ret = append(ret, hrssPublic...)
-	return ret, nil
-}
-
-func (e *cecpq2Curve) accept(rand io.Reader, peerKey []byte) (publicKey []byte, preMasterSecret []byte, err error) {
-	if len(peerKey) != 32+hrss.PublicKeySize {
-		return nil, nil, errors.New("tls: bad length CECPQ2 offer")
-	}
-
-	if _, err := io.ReadFull(rand, e.x25519PrivateKey[:]); err != nil {
-		return nil, nil, err
-	}
-
-	var x25519Shared, x25519PeerKey, x25519Public [32]byte
-	copy(x25519PeerKey[:], peerKey)
-	curve25519.ScalarBaseMult(&x25519Public, &e.x25519PrivateKey)
-	curve25519.ScalarMult(&x25519Shared, &e.x25519PrivateKey, &x25519PeerKey)
-
-	// Per RFC 7748, reject the all-zero value in constant time.
-	var zeros [32]byte
-	if subtle.ConstantTimeCompare(zeros[:], x25519Shared[:]) == 1 {
-		return nil, nil, errors.New("tls: X25519 value with wrong order")
-	}
-
-	hrssPublicKey, ok := hrss.ParsePublicKey(peerKey[32:])
-	if !ok {
-		return nil, nil, errors.New("tls: bad CECPQ2 offer")
-	}
-
-	hrssCiphertext, hrssShared := hrssPublicKey.Encap(rand)
-
-	publicKey = append(publicKey, x25519Public[:]...)
-	publicKey = append(publicKey, hrssCiphertext...)
-	preMasterSecret = append(preMasterSecret, x25519Shared[:]...)
-	preMasterSecret = append(preMasterSecret, hrssShared...)
-
-	return publicKey, preMasterSecret, nil
-}
-
-func (e *cecpq2Curve) finish(peerKey []byte) (preMasterSecret []byte, err error) {
-	if len(peerKey) != 32+hrss.CiphertextSize {
-		return nil, errors.New("tls: bad length CECPQ2 reply")
-	}
-
-	var x25519Shared, x25519PeerKey [32]byte
-	copy(x25519PeerKey[:], peerKey)
-	curve25519.ScalarMult(&x25519Shared, &e.x25519PrivateKey, &x25519PeerKey)
-
-	// Per RFC 7748, reject the all-zero value in constant time.
-	var zeros [32]byte
-	if subtle.ConstantTimeCompare(zeros[:], x25519Shared[:]) == 1 {
-		return nil, errors.New("tls: X25519 value with wrong order")
-	}
-
-	hrssShared, ok := e.hrssPrivateKey.Decap(peerKey[32:])
-	if !ok {
-		return nil, errors.New("tls: invalid HRSS ciphertext")
-	}
-
-	preMasterSecret = append(preMasterSecret, x25519Shared[:]...)
-	preMasterSecret = append(preMasterSecret, hrssShared...)
-
-	return preMasterSecret, nil
-}
-
-func curveForCurveID(id CurveID, config *Config) (ecdhCurve, bool) {
+func kemForCurveID(id CurveID, config *Config) (kemImplementation, bool) {
 	switch id {
 	case CurveP224:
-		return &ellipticECDHCurve{curve: elliptic.P224(), sendCompressed: config.Bugs.SendCompressedCoordinates}, true
+		return &ecdhKEM{curve: elliptic.P224(), sendCompressed: config.Bugs.SendCompressedCoordinates}, true
 	case CurveP256:
-		return &ellipticECDHCurve{curve: elliptic.P256(), sendCompressed: config.Bugs.SendCompressedCoordinates}, true
+		return &ecdhKEM{curve: elliptic.P256(), sendCompressed: config.Bugs.SendCompressedCoordinates}, true
 	case CurveP384:
-		return &ellipticECDHCurve{curve: elliptic.P384(), sendCompressed: config.Bugs.SendCompressedCoordinates}, true
+		return &ecdhKEM{curve: elliptic.P384(), sendCompressed: config.Bugs.SendCompressedCoordinates}, true
 	case CurveP521:
-		return &ellipticECDHCurve{curve: elliptic.P521(), sendCompressed: config.Bugs.SendCompressedCoordinates}, true
+		return &ecdhKEM{curve: elliptic.P521(), sendCompressed: config.Bugs.SendCompressedCoordinates}, true
 	case CurveX25519:
-		return &x25519ECDHCurve{setHighBit: config.Bugs.SetX25519HighBit}, true
-	case CurveCECPQ2:
-		return &cecpq2Curve{}, true
+		return &x25519KEM{setHighBit: config.Bugs.SetX25519HighBit}, true
 	default:
 		return nil, false
 	}
@@ -576,7 +488,7 @@ func (ka *signedKeyAgreement) verifyParameters(config *Config, clientHello *clie
 // either be ECDSA or RSA.
 type ecdheKeyAgreement struct {
 	auth    keyAgreementAuthentication
-	curve   ecdhCurve
+	kem     kemImplementation
 	curveID CurveID
 	peerKey []byte
 }
@@ -588,7 +500,7 @@ func (ka *ecdheKeyAgreement) generateServerKeyExchange(config *Config, cert *Cer
 NextCandidate:
 	for _, candidate := range preferredCurves {
 		if isPqGroup(candidate) && version < VersionTLS13 {
-			// CECPQ2 is TLS 1.3-only.
+			// Post-quantum "groups" require TLS 1.3.
 			continue
 		}
 
@@ -605,12 +517,12 @@ NextCandidate:
 	}
 
 	var ok bool
-	if ka.curve, ok = curveForCurveID(curveid, config); !ok {
+	if ka.kem, ok = kemForCurveID(curveid, config); !ok {
 		return nil, errors.New("tls: preferredCurves includes unsupported curve")
 	}
 	ka.curveID = curveid
 
-	publicKey, err := ka.curve.offer(config.rand())
+	publicKey, err := ka.kem.generate(config.rand())
 	if err != nil {
 		return nil, err
 	}
@@ -636,7 +548,7 @@ func (ka *ecdheKeyAgreement) processClientKeyExchange(config *Config, cert *Cert
 	if len(ckx.ciphertext) == 0 || int(ckx.ciphertext[0]) != len(ckx.ciphertext)-1 {
 		return nil, errClientKeyExchange
 	}
-	return ka.curve.finish(ckx.ciphertext[1:])
+	return ka.kem.decap(ckx.ciphertext[1:])
 }
 
 func (ka *ecdheKeyAgreement) processServerKeyExchange(config *Config, clientHello *clientHelloMsg, serverHello *serverHelloMsg, key crypto.PublicKey, skx *serverKeyExchangeMsg) error {
@@ -646,11 +558,11 @@ func (ka *ecdheKeyAgreement) processServerKeyExchange(config *Config, clientHell
 	if skx.key[0] != 3 { // named curve
 		return errors.New("tls: server selected unsupported curve")
 	}
-	curveid := CurveID(skx.key[1])<<8 | CurveID(skx.key[2])
-	ka.curveID = curveid
+	curveID := CurveID(skx.key[1])<<8 | CurveID(skx.key[2])
+	ka.curveID = curveID
 
 	var ok bool
-	if ka.curve, ok = curveForCurveID(curveid, config); !ok {
+	if ka.kem, ok = kemForCurveID(curveID, config); !ok {
 		return errors.New("tls: server selected unsupported curve")
 	}
 
@@ -668,24 +580,24 @@ func (ka *ecdheKeyAgreement) processServerKeyExchange(config *Config, clientHell
 }
 
 func (ka *ecdheKeyAgreement) generateClientKeyExchange(config *Config, clientHello *clientHelloMsg, cert *x509.Certificate) ([]byte, *clientKeyExchangeMsg, error) {
-	if ka.curve == nil {
+	if ka.kem == nil {
 		return nil, nil, errors.New("missing ServerKeyExchange message")
 	}
 
-	publicKey, preMasterSecret, err := ka.curve.accept(config.rand(), ka.peerKey)
+	ciphertext, secret, err := ka.kem.encap(config.rand(), ka.peerKey)
 	if err != nil {
 		return nil, nil, err
 	}
 
 	ckx := new(clientKeyExchangeMsg)
-	ckx.ciphertext = make([]byte, 1+len(publicKey))
-	ckx.ciphertext[0] = byte(len(publicKey))
-	copy(ckx.ciphertext[1:], publicKey)
+	ckx.ciphertext = make([]byte, 1+len(ciphertext))
+	ckx.ciphertext[0] = byte(len(ciphertext))
+	copy(ckx.ciphertext[1:], ciphertext)
 	if config.Bugs.InvalidECDHPoint {
 		ckx.ciphertext[1] ^= 0xff
 	}
 
-	return preMasterSecret, ckx, nil
+	return secret, ckx, nil
 }
 
 func (ka *ecdheKeyAgreement) peerSignatureAlgorithm() signatureAlgorithm {
